import os
import os.path
import requests
import stat
import sys
import tempfile
from hgsrht.access import UserAccess, get_repo, has_access
from hgsrht.types import Redirect
from srht.config import cfg
from srht.config import cfg
from srht.database import db, DbSession
from werkzeug.wrappers import Request, Response

mercurial_content_type = "application/mercurial"

db = DbSession(cfg("hg.sr.ht", "connection-string"))
db.init()

class HgWebShim:
    """ This is a WSGI application that just runs hgweb directly if
        the request seems to come from command-line `hg`.
    """
    def __init__(self, inner):
        self._inner = inner
        self._repos_dir = cfg("hg.sr.ht", "repos")

        try:
            from mercurial.hgweb import hgweb
            self._hgweb_class = hgweb
        except ImportError:
            print("ERROR: no mercurial installed in Python3's PYTHONPATH!")
            print("ERROR: https cloning won't work!")
            self._hgweb_class = None

    def __call__(self, environ, start_response):
        request = Request(environ)
        # We want hgweb (and, underneath it, mercurial's wire protocol server)
        # to handle this request if it looks vaguely like a repo URL, and
        # has the content-type for mercurial clients.
        if request.headers.get('Accept', '').startswith(mercurial_content_type):
            if self._hgweb_class is None:
                response = Response("wire protocol not available over HTTP",
                                    content_type="application/hg-error",
                                    status=501)
                return response(environ, start_response)

            # Only do something if it's a repo URL (`/username/repo`).
            path_bits = request.path.strip('/').split('/')
            if len(path_bits) != 2:
                response = Response(status=404)
                return response(environ, start_response)

            # Check that this repository is authorized for public https
            # cloning.
            user_name, repo_name = path_bits
            error_resp = self._check_auth(user_name, repo_name)
            if error_resp is not None:
                return error_resp(environ, start_response)

            hgweb_config = self._get_hgweb_config(user_name, repo_name)
            hgweb_app = self._hgweb_class(hgweb_config)
            return hgweb_app(environ, start_response)

        return self._inner(environ, start_response)

    def _check_auth(self, user_name, repo_name):
        owner, repo = get_repo(user_name, repo_name)
        if not repo:
            return Response(status=404)
        if not has_access(repo, UserAccess.read):
            # Use 404 error to not leak information about private repos.
            return Response(status=404)
        if isinstance(repo, Redirect):
            new_user_name = repo.new_repo.owner.canonical_name
            new_repo_name = repo.new_repo.name
            return Response("repository has moved to "
                            f"{new_user_name}/{new_repo_name}",
                            content_type="application/hg-error",
                            status=301)
        # Everything is good, proceed.
        return None

    def _get_hgweb_config(self, user_name, repo_name):
        repo_alias = f"{user_name}/{repo_name}"
        repo_dir = os.path.join(self._repos_dir, repo_alias)
        return {repo_alias.encode(): repo_dir.encode()}
